import os

import matplotlib
from keras.callbacks import Callback

matplotlib.use('Agg')
import matplotlib.pyplot as plt


class VisualizeHistory(Callback):
    def __init__(self, dir_path):
        super(VisualizeHistory, self).__init__()

        self.dir_path = dir_path

    def on_train_begin(self, logs={}):
        self.epoch = []
        self.history = {}

    def on_epoch_end(self, epoch, logs={}):
        self.epoch.append(epoch)
        for k, v in logs.items():
            self.history.setdefault(k, []).append(v)

        for key in filter(lambda k: not k.startswith('val_'), self.history.keys()):
            val_key = 'val_{}'.format(key)

            plt.figure(figsize=(15, 5))

            plt.plot(self.history[key], label='Train', linewidth=3.0)
            if val_key in self.history:
                plt.plot(self.history[val_key], label='Valid', linewidth=3.0)

            plt.legend()
            plt.ylabel(key)
            plt.xlabel('epoch')

            plt.savefig(os.path.join(self.dir_path, '{}.png'.format(key)))
            plt.close()



from rndtools.keras_dataset_generator import DataSerializer
from rndtools.pipeline import Step


class Cache(Step):
    GIVE_ALL = True

    def __init__(self, path):
        self.path = path

    def transform(self, params):
        DataSerializer(self.path).pickle_out(params)


class UnCache(Step):
    GIVE_ALL = True

    def __init__(self, path):
        self.path = path

    def transform(self, params):
        return DataSerializer(self.path).pickle_in()



from abc import abstractmethod
from collections import defaultdict
from operator import itemgetter

import pathos.pools as pp
import six
from six.moves import map
from sklearn.base import BaseEstimator
from tqdm import tqdm


class Step(BaseEstimator):
    GIVE_ALL = False

    @abstractmethod
    def transform(self, params):
        """
        Parameters
        ----------
        params : dict|list
            If GIVE_ALL is False, then param will be dict. It contains all information returned by previous
            steps.
            If GIVE_ALL is True, then param will be list. It contains a dicts that contains all information about all
            examples given to pipeline.

        Returns
        -------
        dict|list|None
            Should return dict or list that contains informations that will be helpful in the consecutive steps
            in a pipeline. Can return None, if you do now pass and informations.

        """
        pass

    def fit_transform(self, params):
        return self.transform(params)


class Pipeline(Step):
    """
    Parameters
    ----------
    show_progressbar : bool, optional
        Flag that indicates if progressbar should be shown.
    parallel : bool, optional
        If parallel is True, then all examples will run in parallel. Default: False.
    """
    def __init__(self, *steps, **kwargs):
        self.show_progressbar = kwargs.get('show_progressbar', True)
        self.parallel = kwargs.get('parallel', False)
        self.steps = steps

    def transform(self, params, f='transform'):
        was_single = False

        if isinstance(params, dict):
            params = [params]
            was_single = True

        result = self._transform_single(params, f)

        if was_single:
            return result[0]

        return result

    def fit_transform(self, params):
        return self.transform(params, f='fit_transform')

    def _transform_single(self, rows, f):
        for name, step in self.steps:
            func = getattr(step, f)

            aa = [(i, row) for i, row in enumerate(rows) if 'error' in row and row['error'] == True]
            indices = map(itemgetter(0), aa)
            for index in sorted(indices, reverse=True):
                del rows[index]

            if step.GIVE_ALL:
                new_rows = func(rows)

                if isinstance(new_rows, list):
                    rows = new_rows
            else:
                if self.parallel:
                    mapper = pp.ProcessPool(4).imap
                else:
                    mapper = map

                mapper = mapper(func, rows)

                if self.show_progressbar:
                    progressbar = tqdm(mapper, total=len(rows), desc=name)
                else:
                    progressbar = mapper

                new_rows = list(progressbar)

                for n, o in zip(new_rows, rows):
                    if isinstance(n, dict):
                        if isinstance(step, Pipeline):
                            n = {"{}_{}".format(name, k):v for k, v in six.iteritems(n)}
                        o.update(n)

            for index, row in sorted(aa, reverse=False, key=itemgetter(0)):
                rows.insert(index, row)

        return rows

    def set_params(self, **params):
        parameters_to_set = defaultdict(dict)
        for k, v in six.iteritems(params):
            transformer, param = k.split('__', 1)
            parameters_to_set[transformer][param] = v

        for name, transformer in self.steps:
            transformer.set_params(**parameters_to_set[name])

    def get_params(self, *args, **kwargs):
        output = {}

        for name, transformer in self.steps:
            for p, v in six.iteritems(transformer.get_params()):
                if isinstance(v, Pipeline):
                    continue

                output['{}__{}'.format(name, p)] = v

        return output



from __future__ import print_function

import inspect
import json
import os
import shutil
from datetime import datetime

import click
import numpy as np
from contexttimer import Timer
from keras.callbacks import CSVLogger
from keras.utils.layer_utils import count_params
from keras.utils import plot_model

from rndtools.io import ensure_dirs
from rndtools.keras_callbacks import VisualizeHistory
from rndtools.keras_dataset_generator import DataSerializer, Dataset


def echo_info(s):
    click.echo('-' * 30)
    click.echo(s)
    click.echo('-' * 30)


class MetaInfo(object):
    def __init__(self):
        self.date = datetime.today()
        self.processing_time = None
        self.number_of_epochs = None
        self.best_test_loss = None
        self.best_train_loss = None
        self.weights_file_size = None
        self.dataset_info = None
        self.number_of_model_parameters = None

    def from_history(self, history):
        self.number_of_epochs = len(history.get('loss'))
        self.best_train_loss = np.round(min(history['loss']), 4)

        if 'val_loss' in history:
            self.best_test_loss = np.round(min(history['val_loss']), 4)

    def serialize(self):
        return {
            'creation_date': str(self.date.date()),
            'processing_time_in_minutes': self.processing_time,
            'number_of_epochs': self.number_of_epochs,
            'best_test_loss': self.best_test_loss,
            'best_train_loss': self.best_train_loss,
            'dataset_info': self.dataset_info,
            'weights_file_size': self.weights_file_size,
            'number_of_model_parameters': self.number_of_model_parameters
        }

    def save_to(self, file):
        json.dump(self.serialize(), open(file, 'w+'), indent=2)


class DatasetInfo(object):
    def __init__(self, data):
        self.data = data

    def serialize(self):
        data = filter(lambda d: isinstance(d, Dataset), self.data)

        return [
            {
                'dir_path': d.path
            }
            for d in data
            ]


def save_summary(f, layers, relevant_nodes=None,
                 line_length=100, positions=[.33, .55, .67, 1.]):
    '''Prints a summary of a layer

    # Arguments
        layers: list of layers to print summaries of
        relevant_nodes: list of relevant nodes
        line_length: total length of printed lines
        positions: relative or absolute positions of log elements in each line
    '''
    if positions[-1] <= 1:
        positions = [int(line_length * p) for p in positions]
    # header names for the different log elements
    to_display = ['Layer (type)', 'Output Shape', 'Param #', 'Connected to']

    def print_row(fields, positions):
        line = ''
        for i in range(len(fields)):
            if i > 0:
                line = line[:-1] + ' '
            line += str(fields[i])
            line = line[:positions[i]]
            line += ' ' * (positions[i] - len(line))
        print(line, file=f)

    print('_' * line_length, file=f)
    print_row(to_display, positions)
    print('=' * line_length, file=f)

    def print_layer_summary(layer):
        try:
            output_shape = layer.output_shape
        except:
            output_shape = 'multiple'
        connections = []
        for node_index, node in enumerate(layer.inbound_nodes):
            if relevant_nodes:
                node_key = layer.name + '_ib-' + str(node_index)
                if node_key not in relevant_nodes:
                    # node is node part of the current network
                    continue
            for i in range(len(node.inbound_layers)):
                inbound_layer = node.inbound_layers[i].name
                inbound_node_index = node.node_indices[i]
                inbound_tensor_index = node.tensor_indices[i]
                connections.append(inbound_layer + '[' + str(inbound_node_index) + '][' + str(inbound_tensor_index) + ']')

        name = layer.name
        cls_name = layer.__class__.__name__
        if not connections:
            first_connection = ''
        else:
            first_connection = connections[0]
        fields = [name + ' (' + cls_name + ')', output_shape, layer.count_params(), first_connection]
        print_row(fields, positions)
        if len(connections) > 1:
            for i in range(1, len(connections)):
                fields = ['', '', '', connections[i]]
                print_row(fields, positions)

    for i in range(len(layers)):
        print_layer_summary(layers[i])
        if i == len(layers) - 1:
            print('=' * line_length, file=f)
        else:
            print('_' * line_length, file=f)

    # trainable_count, non_trainable_count = count_params(layers)
    #
    # print('Total params: {:,}'.format(trainable_count + non_trainable_count), file=f)
    # print('Trainable params: {:,}'.format(trainable_count), file=f)
    # print('Non-trainable params: {:,}'.format(non_trainable_count), file=f)
    # print('_' * line_length, file=f)


def train_model(model_dir, get_model_function, training_function, loading_data_function, overwrite=False):
    """

    :type overwrite: bool
    :type model_dir: str
    :type get_model_function: callable
    :type training_function: callable
    :type loading_data_function: callable
    """

    click.secho('\nModel path: {}'.format(click.format_filename(os.path.abspath(model_dir))), fg='blue')

    if os.path.exists(model_dir):
        if not overwrite:
            click.confirm('\nModel dir already exists. Do you want to REMOVE whole dir before training?', abort=True)

        echo_info('Removing dir...')
        shutil.rmtree(model_dir)

    meta_file_path = os.path.join(model_dir, 'meta.json')
    weights_path = os.path.join(model_dir, 'weights.h5')

    with Timer(factor=1. / 60, output=True, fmt='Process took {:.0f} minutes.') as t:
        meta_info = MetaInfo()

        click.echo()

        echo_info('Creating dirs...')
        ensure_dirs(model_dir)

        echo_info('Creating and compiling model...')
        model = get_model_function()

        echo_info('Saving summary...')
        meta_info.number_of_model_parameters = model.count_params()
        meta_info.save_to(meta_file_path)
        with open(os.path.join(model_dir, 'summary.txt'), 'w+') as f:
            save_summary(f, model.layers)

        echo_info('Saving architecture...')
        open(os.path.join(model_dir, 'architecture.json'), 'w').write(model.to_json())

        echo_info('Plotting model...')
        plot_model(model, to_file=os.path.join(model_dir, 'model.png'), show_shapes=True)

        echo_info('Saving model source code...')
        with open(os.path.join(model_dir, 'model.py'), 'w+') as f:
            f.write(inspect.getsource(get_model_function))
            f.write('\n')
            f.write(inspect.getsource(training_function))

        echo_info('Loading data...')
        data = loading_data_function()
        meta_info.dataset_info = DatasetInfo(data).serialize()
        meta_info.save_to(meta_file_path)

        echo_info('Instantiating callbacks...')
        callbacks = [
            VisualizeHistory(dir_path=model_dir),
            CSVLogger(filename=os.path.join(model_dir, 'history.csv'))
        ]

        echo_info('Training model...')
        with Timer(factor=1. / 60, output=True, fmt='Training model took {:.0f} minutes'):
            params = {
                'data': data,
                'model': model,
                'model_dir': model_dir
            }

            if 'callbacks' in inspect.getargspec(training_function).args:
                params['callbacks'] = callbacks
            else:
                click.secho("'callbacks' argument not found in training function. Callbacks will not be added.", fg='red')

            history = training_function(**params)

        echo_info('Saving history...')
        ds_dataset_out = DataSerializer(os.path.join(model_dir, "history.pkl"))
        ds_dataset_out.pickle_out(history.history)

        echo_info('Saving meta info...')
        meta_info.from_history(history.history)
        meta_info.processing_time = t.elapsed
        if os.path.exists(weights_path): meta_info.weights_file_size = os.stat(weights_path).st_size
        meta_info.save_to(meta_file_path)

        click.secho('\nFinished!', fg='green', bold=True)



import codecs
import itertools
import json
import os
import pickle
from abc import abstractmethod, abstractproperty

import hickle
import numpy as np


class PartsDoNotExist(Exception):
    pass


class DataSerializer():
    def __init__(self, filepath):
        self.fl = filepath

    def pickle_out(self, data):
        out_s = open(self.fl, 'wb')
        pickle.dump(data, out_s, protocol=2)
        out_s.close()

    def hickle_out(self, data):
        hickle.dump(data, self.fl)

    def hickle_in(self):
        return hickle.load(self.fl)

    def json_out(self, data):
        with open(self.fl, 'w') as f:
            json.dump(data, f)

    def pickle_in(self):
        in_s = open(self.fl, 'rb')
        return pickle.load(in_s)

    def json_in(self):
        with codecs.open(self.fl, 'r', encoding='utf8') as f:
            return json.load(f)

    def json_in_chunks(self, chunk_size):
        with open(self.fl) as f:
            out = json.load(f)
            for i in xrange(0, len(out), chunk_size):
                yield out[i:i + chunk_size]


class Dataset(object):
    @abstractproperty
    def path(self):
        pass


class DatasetInPartsGenerator(Dataset):
    def __init__(self, dir_path, batch_size=1):
        self.parts_to_iterate = None

        self.dir_path = dir_path
        self.batch_size = batch_size

        part_files = [os.path.join(dir_path, f) for f in os.listdir(dir_path) if os.path.isfile(os.path.join(dir_path, f))]
        parts_serializers = [DataSerializer(f) for f in part_files]

        if len(parts_serializers) == 0:
            raise PartsDoNotExist()

        self.parts_serializers = itertools.cycle(parts_serializers)

    @property
    def path(self):
        return self.dir_path

    def _change_parts_serializer(self):
        serializer = next(self.parts_serializers)

        X_part, Y_part = serializer.hickle_in()

        parts_to_split = X_part.shape[0] / self.batch_size

        self.parts_to_iterate = itertools.izip(np.array_split(X_part, parts_to_split), np.array_split(Y_part, parts_to_split))

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        try:
            x, y = next(self.parts_to_iterate)
        except (StopIteration, TypeError):
            self._change_parts_serializer()
            x, y = next(self.parts_to_iterate)

        return x, y



import shutil
import tempfile
from unittest import TestCase

import numpy as np
from keras.layers.core import Dense, Activation
from keras.models import Sequential
from keras.optimizers import SGD

from rndtools.train import train_model


class Test(TestCase):
    def test_all_good(self):
        def get_model():
            model = Sequential()
            model.add(Dense(8, input_dim=2))
            model.add(Dense(1))
            model.add(Activation('sigmoid'))

            sgd = SGD(lr=0.1)
            model.compile(loss='binary_crossentropy', optimizer=sgd)

            return model

        def train(model, model_dir, data, callbacks):
            return model.fit(data[0], data[1], show_accuracy=True, batch_size=1, nb_epoch=10, callbacks=callbacks)

        def load_data():
            X = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])
            y = np.array([[0], [1], [1], [0]])

            return X, y

        dirpath = tempfile.mkdtemp()

        train_model(
            dirpath,
            get_model,
            train,
            load_data,
            overwrite=True
        )

        shutil.rmtree(dirpath)



#!/usr/bin/env python
from setuptools import setup, find_packages

install_requires = [
    'numpy >= 1.10',
    'scipy >= 0.11.0',
    'matplotlib >= 1.4',
    'keras >= 2.0',
    'click >= 5.0',
    'contexttimer >= 0.3.3',
    'hickle >= 2.1.0',
    'scikit-learn >= 0.18',
    'six >= 1.11',
    'pathos >= 0.2',
    'tqdm >= 4.19',
]

tests_require = [
    'Tensorflow >= 1.0',
    'mock',
    'nose',
]


def readme():
    return open('index.rst').read()


setup(
    name='rndtools',
    version='3.0.5',
    description='R&D tools',
    long_description=readme(),
    packages=find_packages(),
    install_requires=install_requires,
    tests_require=tests_require,
    test_suite="nose.collector",
    url="https://gitlab.com/identt-rnd/rnd-tools",
    maintainer='Damian Mirecki',
    maintainer_email='damian.mirecki@identt.pl',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)
